@@ -24,14 +24,23 @@
 
 #define GNOME_DESKTOP_USE_UNSTABLE_API
 
+#include <config.h>
+#include <glib/gi18n-lib.h>
 #include <stdlib.h>
 #include <string.h>
 #include <glib.h>
 #include <glib/gstdio.h>
+
+#undef GNOME_DISABLE_DEPRECATED
 #include "libgnomeui/gnome-rr-config.h"
 #include "edid.h"
+#include <X11/Xlib.h>
+#include <X11/extensions/Xrandr.h>
+#include <gdk/gdkx.h>
+#include "gnome-rr-private.h"
 
-#define CONFIG_BASENAME "monitors.xml"
+#define CONFIG_INTENDED_BASENAME "monitors.xml"
+#define CONFIG_BACKUP_BASENAME "monitors.xml.backup"
 
 /* In version 0 of the config file format, we had several <configuration>
  * toplevel elements and no explicit version number.  So, the filed looked
@@ -66,16 +75,16 @@ static gboolean parse_file_gmarkup (const gchar *file,
 
 typedef struct CrtcAssignment CrtcAssignment;
 
-static gboolean         crtc_assignment_apply (CrtcAssignment   *assign);
+static gboolean         crtc_assignment_apply (CrtcAssignment   *assign,
+					       guint32           timestamp,
+					       GError          **error);
 static CrtcAssignment  *crtc_assignment_new   (GnomeRRScreen    *screen,
-					       GnomeOutputInfo **outputs);
+					       GnomeOutputInfo **outputs,
+					       GError          **error);
 static void             crtc_assignment_free  (CrtcAssignment   *assign);
 static void             output_free           (GnomeOutputInfo  *output);
 static GnomeOutputInfo *output_copy           (GnomeOutputInfo  *output);
 
-static gchar *get_old_config_filename (void);
-static gchar *get_config_filename (void);
-
 typedef struct Parser Parser;
 
 /* Parser for monitor configurations */
@@ -180,6 +189,7 @@ handle_start_element (GMarkupParseContext *context,
 	}	
 	parser->output->connected = FALSE;
 	parser->output->on = FALSE;
+	parser->output->primary = FALSE;
     }
     else if (strcmp (name, "configuration") == 0)
     {
@@ -187,7 +197,7 @@ handle_start_element (GMarkupParseContext *context,
 	
 	parser->configuration = g_new0 (GnomeRRConfig, 1);
 	parser->configuration->clone = FALSE;
-	parser->configuration->outputs = g_new0 (GnomeOutputInfo *, 1);
+	parser->configuration->outputs = NULL;
     }
     else if (strcmp (name, "monitors") == 0)
     {
@@ -335,6 +345,13 @@ handle_text (GMarkupParseContext *context,
 	    parser->output->rotation |= GNOME_RR_REFLECT_Y;
 	}
     }
+    else if (stack_is (parser, "primary", "output", "configuration", TOPLEVEL_ELEMENT, NULL))
+    {
+	if (strcmp (text, "yes") == 0)
+	{
+	    parser->output->primary = TRUE;
+	}
+    }
     else
     {
 	/* Ignore other properties so we can expand the format in the future */
@@ -419,42 +436,6 @@ out:
     return result;
 }
 
-static GnomeRRConfig **
-configurations_read (GError **error)
-{
-    char *filename;
-    GnomeRRConfig **configs;
-    GError *err;
-
-    /* Try the new configuration file... */
-
-    filename = get_config_filename ();
-
-    err = NULL;
-    
-    configs = configurations_read_from_file (filename, &err);
-
-    g_free (filename);
-
-    if (err)
-    {
-	if (g_error_matches (err, G_FILE_ERROR, G_FILE_ERROR_NOENT))
-	{
-	    g_error_free (err);
-	    
-	    /* Okay, so try the old configuration file */
-	    filename = get_old_config_filename ();
-	    configs = configurations_read_from_file (filename, error);
-	    g_free (filename);
-	}
-	else
-	{
-	    g_propagate_error (error, err);
-	}
-    }
-    return configs;
-}
-
 GnomeRRConfig *
 gnome_rr_config_new_current (GnomeRRScreen *screen)
 {
@@ -582,6 +563,8 @@ gnome_rr_config_new_current (GnomeRRScreen *screen)
 		output->pref_height = 768;
 	    }
 	}
+
+        output->primary = gnome_rr_output_get_is_primary (rr_output);
  
 	g_ptr_array_add (a, output);
     }
@@ -657,6 +640,8 @@ outputs_free (GnomeOutputInfo **outputs)
 
     for (i = 0; outputs[i] != NULL; ++i)
 	output_free (outputs[i]);
+
+    g_free (outputs);
 }
 
 void
@@ -881,12 +866,19 @@ make_outputs (GnomeRRConfig *config)
 
 gboolean
 gnome_rr_config_applicable (GnomeRRConfig  *configuration,
-			  GnomeRRScreen       *screen)
+			    GnomeRRScreen  *screen,
+			    GError        **error)
 {
-    GnomeOutputInfo **outputs = make_outputs (configuration);
-    CrtcAssignment *assign = crtc_assignment_new (screen, outputs);
+    GnomeOutputInfo **outputs;
+    CrtcAssignment *assign;
     gboolean result;
 
+    g_return_val_if_fail (configuration != NULL, FALSE);
+    g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
+    outputs = make_outputs (configuration);
+    assign = crtc_assignment_new (screen, outputs, error);
+
     if (assign)
     {
 	result = TRUE;
@@ -904,16 +896,24 @@ gnome_rr_config_applicable (GnomeRRConfig  *configuration,
 
 /* Database management */
 
-static gchar *
-get_old_config_filename (void)
+static void
+ensure_config_directory (void)
 {
-    return g_build_filename (g_get_home_dir(), ".gnome2", CONFIG_BASENAME, NULL);
+    g_mkdir_with_parents (g_get_user_config_dir (), 0700);
 }
 
-static gchar *
-get_config_filename (void)
+char *
+gnome_rr_config_get_backup_filename (void)
 {
-    return g_build_filename (g_get_user_config_dir (), CONFIG_BASENAME, NULL);
+    ensure_config_directory ();
+    return g_build_filename (g_get_user_config_dir (), CONFIG_BACKUP_BASENAME, NULL);
+}
+
+char *
+gnome_rr_config_get_intended_filename (void)
+{
+    ensure_config_directory ();
+    return g_build_filename (g_get_user_config_dir (), CONFIG_INTENDED_BASENAME, NULL);
 }
 
 static const char *
@@ -995,6 +995,8 @@ emit_configuration (GnomeRRConfig *config,
 		string, "          <reflect_x>%s</reflect_x>\n", get_reflect_x (output->rotation));
 	    g_string_append_printf (
 		string, "          <reflect_y>%s</reflect_y>\n", get_reflect_y (output->rotation));
+            g_string_append_printf (
+                string, "          <primary>%s</primary>\n", yes_no (output->primary));
 	}
 	
 	g_string_append_printf (string, "      </output>\n");
@@ -1008,6 +1010,7 @@ gnome_rr_config_sanitize (GnomeRRConfig *config)
 {
     int i;
     int x_offset, y_offset;
+    gboolean found;
 
     /* Offset everything by the top/left-most coordinate to
      * make sure the configuration starts at (0, 0)
@@ -1034,19 +1037,45 @@ gnome_rr_config_sanitize (GnomeRRConfig *config)
 	    output->y -= y_offset;
 	}
     }
+
+    /* Only one primary, please */
+    found = FALSE;
+    for (i = 0; config->outputs[i]; ++i)
+    {
+        if (config->outputs[i]->primary)
+        {
+            if (found)
+            {
+                config->outputs[i]->primary = FALSE;
+            }
+            else
+            {
+                found = TRUE;
+            }
+        }
+    }
 }
 
 
 gboolean
-gnome_rr_config_save (GnomeRRConfig *configuration, GError **err)
+gnome_rr_config_save (GnomeRRConfig *configuration, GError **error)
 {
     GnomeRRConfig **configurations;
-    GString *output = g_string_new("");
+    GString *output;
     int i;
-    gchar *filename;
+    gchar *intended_filename;
+    gchar *backup_filename;
     gboolean result;
 
-    configurations = configurations_read (NULL); /* NULL-GError */
+    g_return_val_if_fail (configuration != NULL, FALSE);
+    g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
+    output = g_string_new ("");
+
+    backup_filename = gnome_rr_config_get_backup_filename ();
+    intended_filename = gnome_rr_config_get_intended_filename ();
+
+    configurations = configurations_read_from_file (intended_filename, NULL); /* NULL-GError */
     
     g_string_append_printf (output, "<monitors version=\"1\">\n");
 
@@ -1065,20 +1094,16 @@ gnome_rr_config_save (GnomeRRConfig *configuration, GError **err)
 
     g_string_append_printf (output, "</monitors>\n");
 
-    filename = get_config_filename ();
-    result = g_file_set_contents (filename, output->str, -1, err);
-    g_free (filename);
+    /* backup the file first */
+    rename (intended_filename, backup_filename); /* no error checking because the intended file may not even exist */
 
-    if (result)
-    {
-	/* Only remove the old config file if we were successful in saving the new one */
+    result = g_file_set_contents (intended_filename, output->str, -1, error);
 
-	filename = get_old_config_filename ();
-	if (g_file_test (filename, G_FILE_TEST_EXISTS))
-	    g_unlink (filename);
+    if (!result)
+	rename (backup_filename, intended_filename); /* no error checking because the backup may not even exist */
 
-	g_free (filename);
-    }
+    g_free (backup_filename);
+    g_free (intended_filename);
 
     return result;
 }
@@ -1101,19 +1126,19 @@ gnome_rr_config_copy (GnomeRRConfig *config)
     return copy;
 }
 
-GnomeRRConfig *
-gnome_rr_config_new_stored (GnomeRRScreen *screen)
+static GnomeRRConfig *
+config_new_stored (GnomeRRScreen *screen, const char *filename, GError **error)
 {
     GnomeRRConfig *current;
     GnomeRRConfig **configs;
     GnomeRRConfig *result;
 
-    if (!screen)
-	return NULL;
+    g_return_val_if_fail (screen != NULL, NULL);
+    g_return_val_if_fail (error == NULL || *error == NULL, NULL);
     
     current = gnome_rr_config_new_current (screen);
     
-    configs = configurations_read (NULL); /* NULL_GError */
+    configs = configurations_read_from_file (filename, error);
 
     result = NULL;
     if (configs)
@@ -1129,6 +1154,10 @@ gnome_rr_config_new_stored (GnomeRRScreen *screen)
 	    }
 	}
 
+	if (result == NULL)
+	    g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_NO_MATCHING_CONFIG,
+			 _("none of the saved display configurations matched the active configuration"));
+
 	configurations_free (configs);
     }
 
@@ -1137,9 +1166,36 @@ gnome_rr_config_new_stored (GnomeRRScreen *screen)
     return result;
 }
 
+GnomeRRConfig *
+gnome_rr_config_new_stored (GnomeRRScreen *screen, GError **error)
+{
+    char *intended_filename;
+    GnomeRRConfig *config;
+
+    intended_filename = gnome_rr_config_get_intended_filename ();
+
+    config = config_new_stored (screen, intended_filename, error);
+
+    g_free (intended_filename);
+
+    return config;
+}
+
+#ifndef GNOME_DISABLE_DEPRECATED_SOURCE
 gboolean
 gnome_rr_config_apply (GnomeRRConfig *config,
-		       GnomeRRScreen *screen)
+		       GnomeRRScreen *screen,
+		       GError       **error)
+{
+    return gnome_rr_config_apply_with_time (config, screen, GDK_CURRENT_TIME, error);
+}
+#endif
+
+gboolean
+gnome_rr_config_apply_with_time (GnomeRRConfig *config,
+				 GnomeRRScreen *screen,
+				 guint32        timestamp,
+				 GError       **error)
 {
     CrtcAssignment *assignment;
     GnomeOutputInfo **outputs;
@@ -1147,15 +1203,15 @@ gnome_rr_config_apply (GnomeRRConfig *config,
 
     outputs = make_outputs (config);
 
-    assignment = crtc_assignment_new (screen, outputs);
+    assignment = crtc_assignment_new (screen, outputs, error);
 
     outputs_free (outputs);
     
     if (assignment)
     {
-	if (crtc_assignment_apply (assignment))
+	if (crtc_assignment_apply (assignment, timestamp, error))
 	    result = TRUE;
-	    
+
 	crtc_assignment_free (assignment);
 
 	gdk_flush ();
@@ -1164,25 +1220,137 @@ gnome_rr_config_apply (GnomeRRConfig *config,
     return result;
 }
 
+#ifndef GNOME_DISABLE_DEPRECATED_SOURCE
+/**
+ * gnome_rr_config_apply_stored:
+ * @screen: A #GnomeRRScreen
+ * @error: Location to store error, or %NULL
+ *
+ * See the documentation for gnome_rr_config_apply_from_filename().  This
+ * function simply calls that other function with a filename of
+ * gnome_rr_config_get_intended_filename().
+
+ * @Deprecated: 2.26: Use gnome_rr_config_apply_from_filename() instead and pass it
+ * the filename from gnome_rr_config_get_intended_filename().
+ */
 gboolean
-gnome_rr_config_apply_stored (GnomeRRScreen *screen)
+gnome_rr_config_apply_stored (GnomeRRScreen *screen, GError **error)
+{
+    char *filename;
+    gboolean result;
+
+    filename = gnome_rr_config_get_intended_filename ();
+    result = gnome_rr_config_apply_from_filename_with_time (screen, filename, GDK_CURRENT_TIME, error);
+    g_free (filename);
+
+    return result;
+}
+#endif
+
+#ifndef GNOME_DISABLE_DEPRECATED_SOURCE
+/* gnome_rr_config_apply_from_filename:
+ * @screen: A #GnomeRRScreen
+ * @filename: Path of the file to look in for stored RANDR configurations.
+ * @error: Location to store error, or %NULL
+ *
+ * First, this function refreshes the @screen to match the current RANDR
+ * configuration from the X server.  Then, it tries to load the file in
+ * @filename and looks for suitable matching RANDR configurations in the file;
+ * if one is found, that configuration will be applied to the current set of
+ * RANDR outputs.
+ *
+ * Typically, @filename is the result of gnome_rr_config_get_intended_filename() or
+ * gnome_rr_config_get_backup_filename().
+ *
+ * Returns: TRUE if the RANDR configuration was loaded and applied from
+ * $(XDG_CONFIG_HOME)/monitors.xml, or FALSE otherwise:
+ *
+ * If the current RANDR configuration could not be refreshed, the @error will
+ * have a domain of #GNOME_RR_ERROR and a corresponding error code.
+ *
+ * If the file in question is loaded successfully but the configuration cannot
+ * be applied, the @error will have a domain of #GNOME_RR_ERROR.  Note that an
+ * error code of #GNOME_RR_ERROR_NO_MATCHING_CONFIG is not a real error; it
+ * simply means that there were no stored configurations that match the current
+ * set of RANDR outputs.
+ *
+ * If the file in question cannot be loaded, the @error will have a domain of
+ * #G_FILE_ERROR.  Note that an error code of G_FILE_ERROR_NOENT is not really
+ * an error, either; it means that there was no stored configuration file and so
+ * nothing is changed.
+ *
+ * @Deprecated: 2.28: use gnome_rr_config_apply_from_filename_with_time() instead.
+ */
+gboolean
+gnome_rr_config_apply_from_filename (GnomeRRScreen *screen, const char *filename, GError **error)
+{
+    return gnome_rr_config_apply_from_filename_with_time (screen, filename, GDK_CURRENT_TIME, error);
+}
+#endif
+
+/* gnome_rr_config_apply_from_filename_with_time:
+ * @screen: A #GnomeRRScreen
+ * @filename: Path of the file to look in for stored RANDR configurations.
+ * @timestamp: X server timestamp from the event that causes the screen configuration to change (a user's button press, for example)
+ * @error: Location to store error, or %NULL
+ *
+ * First, this function refreshes the @screen to match the current RANDR
+ * configuration from the X server.  Then, it tries to load the file in
+ * @filename and looks for suitable matching RANDR configurations in the file;
+ * if one is found, that configuration will be applied to the current set of
+ * RANDR outputs.
+ *
+ * Typically, @filename is the result of gnome_rr_config_get_intended_filename() or
+ * gnome_rr_config_get_backup_filename().
+ *
+ * Returns: TRUE if the RANDR configuration was loaded and applied from
+ * $(XDG_CONFIG_HOME)/monitors.xml, or FALSE otherwise:
+ *
+ * If the current RANDR configuration could not be refreshed, the @error will
+ * have a domain of #GNOME_RR_ERROR and a corresponding error code.
+ *
+ * If the file in question is loaded successfully but the configuration cannot
+ * be applied, the @error will have a domain of #GNOME_RR_ERROR.  Note that an
+ * error code of #GNOME_RR_ERROR_NO_MATCHING_CONFIG is not a real error; it
+ * simply means that there were no stored configurations that match the current
+ * set of RANDR outputs.
+ *
+ * If the file in question cannot be loaded, the @error will have a domain of
+ * #G_FILE_ERROR.  Note that an error code of G_FILE_ERROR_NOENT is not really
+ * an error, either; it means that there was no stored configuration file and so
+ * nothing is changed.
+ */
+gboolean
+gnome_rr_config_apply_from_filename_with_time (GnomeRRScreen *screen, const char *filename, guint32 timestamp, GError **error)
 {
     GnomeRRConfig *stored;
+    GError *my_error;
 
-    if (!screen)
-	return FALSE;
+    g_return_val_if_fail (screen != NULL, FALSE);
+    g_return_val_if_fail (filename != NULL, FALSE);
+    g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
-    gnome_rr_screen_refresh (screen);
+    my_error = NULL;
+    if (!gnome_rr_screen_refresh (screen, &my_error)) {
+	    if (my_error) {
+		    g_propagate_error (error, my_error);
+		    return FALSE; /* This is a genuine error */
+	    }
 
-    stored = gnome_rr_config_new_stored (screen);
+	    /* This means the screen didn't change, so just proceed */
+    }
+
+    stored = config_new_stored (screen, filename, error);
 
     if (stored)
     {
-	gnome_rr_config_apply (stored, screen);
+	gboolean result;
+
+	result = gnome_rr_config_apply_with_time (stored, screen, timestamp, error);
 
 	gnome_rr_config_free (stored);
 	
-	return TRUE;
+	return result;
     }
     else
     {
@@ -1190,28 +1358,6 @@ gnome_rr_config_apply_stored (GnomeRRScreen *screen)
     }
 }
 
-static gboolean
-has_similar_mode (GnomeRROutput *output, GnomeRRMode *mode)
-{
-    int i;
-    GnomeRRMode **modes = gnome_rr_output_list_modes (output);
-    int width = gnome_rr_mode_get_width (mode);
-    int height = gnome_rr_mode_get_height (mode);
-
-    for (i = 0; modes[i] != NULL; ++i)
-    {
-	GnomeRRMode *m = modes[i];
-
-	if (gnome_rr_mode_get_width (m) == width	&&
-	    gnome_rr_mode_get_height (m) == height)
-	{
-	    return TRUE;
-	}
-    }
-
-    return FALSE;
-}
-
 /*
  * CRTC assignment
  */
@@ -1230,6 +1376,7 @@ struct CrtcAssignment
 {
     GnomeRRScreen *screen;
     GHashTable *info;
+    GnomeRROutput *primary;
 };
 
 static gboolean
@@ -1256,6 +1403,7 @@ crtc_assignment_assign (CrtcAssignment   *assign,
 			int               x,
 			int               y,
 			GnomeRRRotation   rotation,
+                        gboolean          primary,
 			GnomeRROutput    *output)
 {
     CrtcInfo *info = g_hash_table_lookup (assign->info, crtc);
@@ -1277,6 +1425,11 @@ crtc_assignment_assign (CrtcAssignment   *assign,
 	{
 	    g_ptr_array_add (info->outputs, output);
 
+            if (primary && !assign->primary)
+            {
+                assign->primary = output;
+            }
+
 	    return TRUE;
 	}
     }
@@ -1294,6 +1447,11 @@ crtc_assignment_assign (CrtcAssignment   *assign,
 	
 	g_hash_table_insert (assign->info, crtc, info);
 	    
+        if (primary && !assign->primary)
+        {
+            assign->primary = output;
+        }
+
 	return TRUE;
     }
     
@@ -1311,6 +1469,11 @@ crtc_assignment_unassign (CrtcAssignment *assign,
     {
 	g_ptr_array_remove (info->outputs, output);
 
+        if (assign->primary == output)
+        {
+            assign->primary = NULL;
+        }
+
 	if (info->outputs->len == 0)
 	    g_hash_table_remove (assign->info, crtc);
     }
@@ -1324,6 +1487,12 @@ crtc_assignment_free (CrtcAssignment *assign)
     g_free (assign);
 }
 
+typedef struct {
+    guint32 timestamp;
+    gboolean has_error;
+    GError **error;
+} ConfigureCrtcState;
+
 static void
 configure_crtc (gpointer key,
 		gpointer value,
@@ -1331,13 +1500,20 @@ configure_crtc (gpointer key,
 {
     GnomeRRCrtc *crtc = key;
     CrtcInfo *info = value;
+    ConfigureCrtcState *state = data;
 
-    gnome_rr_crtc_set_config (crtc,
-			info->x, info->y,
-			info->mode,
-			info->rotation,
-			(GnomeRROutput **)info->outputs->pdata,
-			info->outputs->len);
+    if (state->has_error)
+	return;
+
+    if (!gnome_rr_crtc_set_config_with_time (crtc,
+					     state->timestamp,
+					     info->x, info->y,
+					     info->mode,
+					     info->rotation,
+					     (GnomeRROutput **)info->outputs->pdata,
+					     info->outputs->len,
+					     state->error))
+	state->has_error = TRUE;
 }
 
 static gboolean
@@ -1418,6 +1594,7 @@ real_assign_crtcs (GnomeRRScreen *screen,
 			    assignment, crtc, modes[j],
 			    output->x, output->y,
 			    output->rotation,
+                            output->primary,
 			    gnome_rr_output))
 		    {
 			if (real_assign_crtcs (screen, outputs + 1, assignment))
@@ -1478,7 +1655,7 @@ get_required_virtual_size (CrtcAssignment *assign, int *width, int *height)
 }
 
 static CrtcAssignment *
-crtc_assignment_new (GnomeRRScreen *screen, GnomeOutputInfo **outputs)
+crtc_assignment_new (GnomeRRScreen *screen, GnomeOutputInfo **outputs, GError **error)
 {
     CrtcAssignment *assignment = g_new0 (CrtcAssignment, 1);
 
@@ -1498,6 +1675,15 @@ crtc_assignment_new (GnomeRRScreen *screen, GnomeOutputInfo **outputs)
 	if (width < min_width || width > max_width ||
 	    height < min_height || height > max_height)
 	{
+	    g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_BOUNDS_ERROR,
+			 /* Translators: the "requested", "minimum", and
+			  * "maximum" words here are not keywords; please
+			  * translate them as usual. */
+			 _("required virtual size does not fit available size: "
+			   "requested=(%d, %d), minimum=(%d, %d), maximum=(%d, %d)"),
+			 width, height,
+			 min_width, min_height,
+			 max_width, max_height);
 	    goto fail;
 	}
 
@@ -1505,6 +1691,9 @@ crtc_assignment_new (GnomeRRScreen *screen, GnomeOutputInfo **outputs)
 	
 	return assignment;
     }
+    else
+	g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_CRTC_ASSIGNMENT,
+		     _("could not find a suitable configuration of screens"));
     
 fail:
     crtc_assignment_free (assignment);
@@ -1513,7 +1702,7 @@ fail:
 }
 
 static gboolean
-crtc_assignment_apply (CrtcAssignment *assign)
+crtc_assignment_apply (CrtcAssignment *assign, guint32 timestamp, GError **error)
 {
     GnomeRRCrtc **all_crtcs = gnome_rr_screen_list_crtcs (assign->screen);
     int width, height;
@@ -1535,7 +1724,16 @@ crtc_assignment_apply (CrtcAssignment *assign)
     width = MIN (max_width, width);
     height = MAX (min_height, height);
     height = MIN (max_height, height);
-    
+
+    /* FMQ: do we need to check the sizes instead of clamping them? */
+
+    /* Grab the server while we fiddle with the CRTCs and the screen, so that
+     * apps that listen for RANDR notifications will only receive the final
+     * status.
+     */
+
+    gdk_x11_display_grab (gdk_screen_get_display (assign->screen->gdk_screen));
+
     /* Turn off all crtcs that are currently displaying outside the new screen,
      * or are not used in the new setup
      */
@@ -1552,7 +1750,7 @@ crtc_assignment_apply (CrtcAssignment *assign)
 
 	    w = gnome_rr_mode_get_width (mode);
 	    h = gnome_rr_mode_get_height (mode);
-	    
+
 	    if (crtc_is_rotated (crtc))
 	    {
 		int tmp = h;
@@ -1562,7 +1760,7 @@ crtc_assignment_apply (CrtcAssignment *assign)
 	    
 	    if (x + w > width || y + h > height || !g_hash_table_lookup (assign->info, crtc))
 	    {
-		if (!gnome_rr_crtc_set_config (crtc, 0, 0, NULL, GNOME_RR_ROTATION_0, NULL, 0))
+		if (!gnome_rr_crtc_set_config_with_time (crtc, timestamp, 0, 0, NULL, GNOME_RR_ROTATION_0, NULL, 0, error))
 		{
 		    success = FALSE;
 		    break;
@@ -1583,10 +1781,22 @@ crtc_assignment_apply (CrtcAssignment *assign)
 
     if (success)
     {
+	ConfigureCrtcState state;
+
 	gnome_rr_screen_set_size (assign->screen, width, height, width_mm, height_mm);
+
+	state.timestamp = timestamp;
+	state.has_error = FALSE;
+	state.error = error;
 	
-	g_hash_table_foreach (assign->info, configure_crtc, NULL);
+	g_hash_table_foreach (assign->info, configure_crtc, &state);
+
+	success = !state.has_error;
     }
 
+    gnome_rr_screen_set_primary_output (assign->screen, assign->primary);
+
+    gdk_x11_display_ungrab (gdk_screen_get_display (assign->screen->gdk_screen));
+
     return success;
 }
